/*
 * Copyright (C) 2019 by Sukchan Lee <acetcom@gmail.com>
 *
 * This file is part of Open5GS.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "sgw-context.h"
#include "sgw-sm.h"
#include "sgw-event.h"

static ogs_thread_t *thread;
static void sgw_main(void *data);

static int initialized = 0;

int sgw_initialize()
{
    int rv;

    sgw_context_init();
    sgw_event_init();

    rv = ogs_gtp_xact_init(sgw_self()->timer_mgr, 512);
    if (rv != OGS_OK) return rv;

    rv = sgw_context_parse_config();
    if (rv != OGS_OK) return rv;

    rv = ogs_log_config_domain(
            ogs_config()->logger.domain, ogs_config()->logger.level);
    if (rv != OGS_OK) return rv;

    thread = ogs_thread_create(sgw_main, NULL);
    if (!thread) return OGS_ERROR;

    initialized = 1;

    return OGS_OK;
}

void sgw_terminate(void)
{
    if (!initialized) return;

    sgw_event_term();

    ogs_thread_destroy(thread);

    sgw_context_final();

    ogs_gtp_xact_final();

    sgw_event_final();
}

static void sgw_main(void *data)
{
    ogs_fsm_t sgw_sm;
    int rv;

    ogs_fsm_create(&sgw_sm, sgw_state_initial, sgw_state_final);
    ogs_fsm_init(&sgw_sm, 0);

    for ( ;; ) {
        ogs_pollset_poll(sgw_self()->pollset,
                ogs_timer_mgr_next(sgw_self()->timer_mgr));

        /* Process the MESSAGE FIRST.
         *
         * For example, if UE Context Release Complete is received,
         * the MME_TIMER_UE_CONTEXT_RELEASE is first stopped */
        for ( ;; ) {
            sgw_event_t *e = NULL;

            rv = ogs_queue_trypop(sgw_self()->queue, (void**)&e);
            ogs_assert(rv != OGS_ERROR);

            if (rv == OGS_DONE)
                goto done;

            if (rv == OGS_RETRY)
                break;

            ogs_assert(e);
            ogs_fsm_dispatch(&sgw_sm, e);
            sgw_event_free(e);
        }

        ogs_timer_mgr_expire(sgw_self()->timer_mgr);

        /* AND THEN, process the TIMER. */
        for ( ;; ) {
            sgw_event_t *e = NULL;

            rv = ogs_queue_trypop(sgw_self()->queue, (void**)&e);
            ogs_assert(rv != OGS_ERROR);

            if (rv == OGS_DONE)
                goto done;

            if (rv == OGS_RETRY)
                break;

            ogs_assert(e);
            ogs_fsm_dispatch(&sgw_sm, e);
            sgw_event_free(e);
        }
    }
done:

    ogs_fsm_fini(&sgw_sm, 0);
    ogs_fsm_delete(&sgw_sm);
}
